
__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import errno
import shutil

import radical.utils as ru

from ...  import utils     as rpu
from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentStagingOutputComponent

from ...staging_directives import complete_url


# ------------------------------------------------------------------------------
#
class Default(AgentStagingOutputComponent):
    """
    This component performs all agent side output staging directives for compute
    tasks.  It gets tasks from the agent_staging_output_queue, in
    AGENT_STAGING_OUTPUT_PENDING state, will advance them to
    AGENT_STAGING_OUTPUT state while performing the staging, and then moves then
    to the TMGR_STAGING_OUTPUT_PENDING state, which at the moment requires the
    state change to be published to MongoDB (no push into a queue).

    Note that this component also collects stdout/stderr of the tasks (which
    can also be considered staging, really).
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentStagingOutputComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()
        self._stager = rpu.StagingHelper(log=self._log)

        self.register_input(rps.AGENT_STAGING_OUTPUT_PENDING,
                            rpc.AGENT_STAGING_OUTPUT_QUEUE, self.work)

        self.register_output(rps.TMGR_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_COLLECTING_QUEUE)



    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance(tasks, rps.AGENT_STAGING_OUTPUT, publish=True, push=False)

        # we first filter out any tasks which don't need any input staging, and
        # advance them again as a bulk.  We work over the others one by one, and
        # advance them individually, to avoid stalling from slow staging ops.

        no_staging_tasks = list()
        staging_tasks    = list()

        for task in ru.as_list(tasks):

            try:
                uid = task['uid']

                # From here on, any state update will hand control over to the
                # tmgr again.  The next task update should thus push *all* task
                # details, not only state.
                task['$all']    = True
                task['control'] = 'tmgr_pending'

                self._log.debug('=== staging task %s: %s', uid, task['state'])
                import pprint
                self._log.debug('=== staging task %s: %s', uid, pprint.pformat(task))

                # we always dig for stdout/stderr
                self._handle_task_stdio(task)

                # NOTE: all tasks get here after execution, even those which did not
                #       finish successfully.  We do that so that we can make
                #       stdout/stderr available for failed tasks (see
                #       _handle_task_stdio above).  But we don't need to perform any
                #       other staging for those tasks, and in fact can make them
                #       final.
                if task['target_state'] != rps.DONE \
                        and not task['description'].get('stage_on_error'):
                    task['state'] = task['target_state']
                    self._log.debug('task %s skips staging: %s', uid, task['state'])
                    no_staging_tasks.append(task)
                    continue

                # check if we have any staging directives to be enacted in this
                # component
                actionables = list()
                for sd in task['description'].get('output_staging', []):
                    if sd['action'] in [rpc.LINK, rpc.COPY, rpc.MOVE]:
                        actionables.append(sd)

                if actionables:
                    # this task needs some staging
                    staging_tasks.append([task, actionables])
                else:
                    # this task does not need any staging at this point, and can be
                    # advanced
                    task['state'] = rps.TMGR_STAGING_OUTPUT_PENDING
                    no_staging_tasks.append(task)

            except Exception as e:
                self._log.exception('staging prep error')
                task['exception']        = repr(e)
                task['exception_detail'] = '\n'.join(ru.get_exception_trace())

                self.advance(task, rps.FAILED)


        if no_staging_tasks:
            self.advance(no_staging_tasks, rps.TMGR_STAGING_OUTPUT_PENDING,
                                           publish=True, push=True)

        for task, actionables in staging_tasks:
            try:
                self._handle_task_staging(task, actionables)

            except Exception as e:
                self._log.exception('staging error')
                task['exception']        = repr(e)
                task['exception_detail'] = '\n'.join(ru.get_exception_trace())

                self.advance(task, rps.FAILED)


    # --------------------------------------------------------------------------
    #
    def _handle_task_stdio(self, task):

        sbox = task.get('task_sandbox_path')
        uid  = task['uid']

        self._log.debug('out %s: %s', uid, sbox)

        self._prof.prof('staging_stdout_start', uid=uid)
        self._log.debug('out %s: %s', uid, task.get('stdout_file'))

        # TODO: disable this at scale?
        if task.get('stdout_file') and os.path.isfile(task['stdout_file']):
            with ru.ru_open(task['stdout_file'], 'r', errors='ignore') \
                 as stdout_f:
                try:
                    txt = ru.as_string(stdout_f.read())
                except UnicodeDecodeError:
                    txt = "task stdout is binary -- use file staging"

                task['stdout'] += rpu.tail(txt)

        self._prof.prof('staging_stdout_stop',  uid=uid)
        self._prof.prof('staging_stderr_start', uid=uid)

        # TODO: disable this at scale?
        if task.get('stderr_file') and os.path.isfile(task['stderr_file']):
            with ru.ru_open(task['stderr_file'], 'r', errors='ignore') \
                 as stderr_f:
                try:
                    txt = ru.as_string(stderr_f.read())
                except UnicodeDecodeError:
                    txt = "task stderr is binary -- use file staging"

                task['stderr'] += rpu.tail(txt)

        self._prof.prof('staging_stderr_stop', uid=uid)
        self._prof.prof('staging_uprof_start', uid=uid)

      # task_prof = '%s/%s.prof' % (sbox, uid)
      # if os.path.isfile(task_prof):
      #     pids = {}
      #     try:
      #         with ru.ru_open(task_prof, 'r', errors='ignore') as prof_f:
      #             txt = ru.as_string(prof_f.read())
      #             for line in txt.split("\n"):
      #                 if not line:
      #                     continue
      #                 if line[0] == '#':
      #                     continue
      #                 ts, event, comp, tid, _uid, state, msg = \
      #                                                      line.split(',')
      #                 self._prof.prof(ts=float(ts), event=event,
      #                                 comp=comp, tid=tid, uid=_uid,
      #                                 state=state, msg=msg)
      #                 # collect task related PIDs
      #                 if 'RP_LAUNCH_PID' in msg:
      #                     pids['launch_pid'] = int(msg.split('=')[1])
      #                 elif 'RP_RANK_PID' in msg:
      #                     epid_msg, rpid_msg = msg.split(':')
      #                     pids.setdefault('exec_pid', []) \
      #                         .append(int(epid_msg.split('=')[1]))
      #                     pids.setdefault('rank_pid', [])\
      #                         .append(int(rpid_msg.split('=')[1]))
      #     except Exception as e:
      #         self._log.error("Pre/Post profile read failed: `%s`" % e)
      #
      #     if pids:
      #         # keep process IDs within metadata in the task description
      #         if not task['description'].get('metadata'):
      #             task['description']['metadata'] = {}
      #         task['description']['metadata'].update(pids)

        self._prof.prof('staging_uprof_stop', uid=uid)
        self._prof.prof('staging_ofile_start', uid=uid)

        task_ofile = '%s/%s.ofiles' % (sbox, uid)
        if os.path.isfile(task_ofile):
            try:
                with ru.ru_open(task_ofile, 'r', errors='ignore') as fin:
                    task['ofiles'] = [l.strip() for l in fin.readlines()]
            except Exception as e:
                self._log.error("Pre/Post ofile read failed: `%s`" % e)

        self._prof.prof('staging_ofile_stop', uid=uid)


    # --------------------------------------------------------------------------
    #
    def _handle_task_staging(self, task, actionables):

        uid = task['uid']

        # By definition, this compoentn lives on the pilot's target resource.
        # As such, we *know* that all staging ops which would refer to the
        # resource now refer to file://localhost, and thus translate the task,
        # pilot and resource sandboxes into that scope.  Some assumptions are
        # made though:
        #
        #   * paths are directly translatable across schemas
        #   * resource level storage is in fact accessible via file://
        #
        # FIXME: this is costly and should be cached.

        task_sandbox     = ru.Url(task['task_sandbox'])
        pilot_sandbox    = ru.Url(task['pilot_sandbox'])
        session_sandbox  = ru.Url(task['session_sandbox'])
        resource_sandbox = ru.Url(task['resource_sandbox'])
        endpoint_fs      = ru.Url(task['endpoint_fs'])

        task_sandbox.schema     = 'file'
        pilot_sandbox.schema    = 'file'
        session_sandbox.schema  = 'file'
        resource_sandbox.schema = 'file'
        endpoint_fs.schema      = 'file'

        task_sandbox.host       = 'localhost'
        pilot_sandbox.host      = 'localhost'
        session_sandbox.host    = 'localhost'
        resource_sandbox.host   = 'localhost'
        endpoint_fs.host        = 'localhost'

        src_context = {'pwd'      : str(task_sandbox),       # !!!
                       'task'     : str(task_sandbox),
                       'pilot'    : str(pilot_sandbox),
                       'session'  : str(session_sandbox),
                       'resource' : str(resource_sandbox),
                       'endpoint' : str(endpoint_fs)}
        tgt_context = {'pwd'      : str(task_sandbox),       # !!!
                       'task'     : str(task_sandbox),
                       'pilot'    : str(pilot_sandbox),
                       'session'  : str(session_sandbox),
                       'resource' : str(resource_sandbox),
                       'endpoint' : str(endpoint_fs)}


        # we can now handle the actionable staging directives
        for sd in actionables:

            action = sd['action']
            flags  = sd['flags']
            did    = sd['uid']
            src    = sd['source']
            tgt    = sd['target']

            self._prof.prof('staging_out_start', uid=uid, msg=did)

            # agent stager only handles local actions
            if action not in [rpc.COPY, rpc.LINK, rpc.MOVE]:
                self._prof.prof('staging_in_skip', uid=uid, msg=did)
                continue

            # we assume current directory is the task staging 'task://'
            # and we assume the file to be copied is the base filename
            # of the source
            if tgt is None: tgt = ''
            if tgt.strip() == '':
                tgt = 'task:///{}'.format(os.path.basename(src))
            # Fix for when the target PATH is exists *and* it is a folder
            # we assume the 'current directory' is the target folder
            # and we assume the file to be copied is the base filename
            # of the source
            elif os.path.exists(tgt.strip()) and os.path.isdir(tgt.strip()):
                tgt = os.path.join(tgt, os.path.basename(src))

            src = complete_url(src, src_context, self._log)
            tgt = complete_url(tgt, tgt_context, self._log)

            # Currently, we use the same schema for files and folders.
            assert src.schema == 'file', 'staging src must be file://'

            if action in [rpc.COPY, rpc.LINK, rpc.MOVE]:
                assert tgt.schema == 'file', 'staging tgt expected as file://'

            self._stager.handle_staging_directive({'source': src,
                                                   'target': tgt,
                                                   'action': action,
                                                   'flags' : flags})


            self._prof.prof('staging_out_stop', uid=uid, msg=did)

        # all agent staging is done -- pass on to tmgr output staging
        self.advance(task, rps.TMGR_STAGING_OUTPUT_PENDING,
                           publish=True, push=True)


# ------------------------------------------------------------------------------





# pylint: disable=protected-access

__copyright__ = 'Copyright 2016-2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import time
import signal

import radical.utils as ru


# 'enum' for launch method types
LM_NAME_APRUN         = 'APRUN'
LM_NAME_CCMRUN        = 'CCMRUN'
LM_NAME_FORK          = 'FORK'
LM_NAME_IBRUN         = 'IBRUN'
LM_NAME_DRAGON        = 'DRAGON'
LM_NAME_MPIEXEC       = 'MPIEXEC'
LM_NAME_MPIEXEC_MPT   = 'MPIEXEC_MPT'
LM_NAME_MPIRUN        = 'MPIRUN'
LM_NAME_MPIRUN_MPT    = 'MPIRUN_MPT'
LM_NAME_MPIRUN_CCMRUN = 'MPIRUN_CCMRUN'
LM_NAME_MPIRUN_DPLACE = 'MPIRUN_DPLACE'
LM_NAME_MPIRUN_RSH    = 'MPIRUN_RSH'
LM_NAME_JSRUN         = 'JSRUN'
LM_NAME_JSRUN_ERF     = 'JSRUN_ERF'
LM_NAME_PRTE          = 'PRTE'
LM_NAME_FLUX          = 'FLUX'
LM_NAME_RSH           = 'RSH'
LM_NAME_SSH           = 'SSH'
LM_NAME_SRUN          = 'SRUN'

PWD = os.getcwd()


# FIXME: add a well defined LMInfo TypedDict, similar to RMInfo


# ------------------------------------------------------------------------------
#
class LaunchMethod(object):

    MPI_FLAVOR_OMPI     = 'OMPI'
    MPI_FLAVOR_HYDRA    = 'HYDRA'
    MPI_FLAVOR_SPECTRUM = 'SPECTRUM'
    MPI_FLAVOR_PALS     = 'PALS'
    MPI_FLAVOR_UNKNOWN  = 'unknown'

    LM_INVALID          = 'invalid'
    LM_EMPTY            = 'empty'

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self.name       = name
        self._lm_cfg    = lm_cfg
        self._rm_info   = rm_info
        self._log       = log
        self._prof      = prof
        self._pwd       = os.getcwd()
        self._env_orig  = ru.env_eval('env/bs0_orig.env')
        self._in_pytest = False

        reg     = ru.zmq.RegistryClient(url=self._lm_cfg.reg_addr)
        lm_info = reg.get('lm.%s' % self.name.lower())

        self._log.debug('initialize LM %s', self.name)

        if lm_info == self.LM_INVALID:
            self._log.warn('LM info invalid - skip %s', lm_info)

        elif lm_info == self.LM_EMPTY:
            self._log.info('LM info empty for %s', lm_info)

        elif not lm_info:

            # The registry does not yet contain any info for this LM - we need
            # to initialize the LM from scratch.  That happens in the env
            # defined by lm_cfg (derived from the original bs0 env)
            env_sh   = 'env/lm_%s.sh' % self.name.lower()
            env_lm   = ru.env_prep(environment=self._env_orig,
                          pre_exec=lm_cfg.get('pre_exec'),
                          pre_exec_cached=lm_cfg.get('pre_exec_cached'),
                          script_path=env_sh)

            lm_info = self._init_from_scratch(env_lm, env_sh)

            # store the info in the registry for any other instances of the LM
            reg.put('lm.%s' % self.name.lower(), lm_info)

        if lm_info != self.LM_INVALID:
            self.init_from_info(lm_info)

        reg.close()


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Launch Method.
    #
    @classmethod
    def create(cls, name, lm_cfg, rm_info, log, prof):

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod create only available to base class!")

        # In case of undefined LM just return None
        if not name:
            return None

        from .aprun          import APRun
        from .ccmrun         import CCMRun
        from .fork           import Fork
        from .ibrun          import IBRun
        from .dragon         import Dragon
        from .mpiexec        import MPIExec
        from .mpirun         import MPIRun
        from .jsrun          import JSRUN
        from .prte           import PRTE
        from .flux           import Flux
        from .rsh            import RSH
        from .ssh            import SSH
        from .srun           import Srun

        impl = {
            LM_NAME_APRUN         : APRun,
            LM_NAME_CCMRUN        : CCMRun,
            LM_NAME_DRAGON        : Dragon,
            LM_NAME_FORK          : Fork,
            LM_NAME_FLUX          : Flux,
            LM_NAME_IBRUN         : IBRun,
            LM_NAME_MPIEXEC       : MPIExec,
            LM_NAME_MPIEXEC_MPT   : MPIExec,
            LM_NAME_MPIRUN        : MPIRun,
            LM_NAME_MPIRUN_CCMRUN : MPIRun,
            LM_NAME_MPIRUN_RSH    : MPIRun,
            LM_NAME_MPIRUN_MPT    : MPIRun,
            LM_NAME_MPIRUN_DPLACE : MPIRun,
            LM_NAME_JSRUN         : JSRUN,
            LM_NAME_JSRUN_ERF     : JSRUN,
            LM_NAME_PRTE          : PRTE,
            LM_NAME_RSH           : RSH,
            LM_NAME_SSH           : SSH,
            LM_NAME_SRUN          : Srun,
        }

        if name not in impl:
            raise ValueError('LaunchMethod %s unknown' % name)

        return impl[name](name, lm_cfg, rm_info, log, prof)


    # --------------------------------------------------------------------------
    def _init_from_scratch(self, env, env_sh):

        lm_info = None

        # run init_from_scratch in a process under the LM env
        try:
            self._envp = ru.EnvProcess(env=env)
            with self._envp:
                if self._envp:
                    try:
                        data = self.init_from_scratch(env, env_sh)
                        self._envp.put(data)
                    except:
                        self._log.exception('LM init failed')
                        raise

            lm_info = self._envp.get()

        except Exception:
            self._log.warn('LM init failed')
            lm_info = self.LM_INVALID

        # force non-empty lm_info
        if not lm_info:
            self._log.warn('LM init came up empty')
            lm_info = self.LM_EMPTY

        return lm_info


    # --------------------------------------------------------------------------
    #
    def init_from_scratch(self, env, env_sh):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def init_from_info(self, lm_info):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, task, pid):
        '''
        This method cancels the task, in the default case by killing the task's
        launch process identified by its process ID.
        '''

        # kill the process group which should include the actual launch method
        try:
            self._log.debug('killing task %s (%d)', task['uid'], pid)
            os.killpg(pid, signal.SIGTERM)

            # also send a SIGKILL to drive the message home.
            # NOTE: the `sleep` will limit the cancel throughput!
            try:
                time.sleep(0.1)
                os.killpg(pid, signal.SIGKILL)
            except OSError:
                pass

        except OSError:
            # lost race: task is already gone, we ignore this
            self._log.debug('task already gone: %s', task['uid'])


    # --------------------------------------------------------------------------
    #
    def get_task_named_env(self, env_name):

        # we assume that the launcher env is still active in the task execution
        # script.  We thus remove the launcher env from the task env before
        # applying the task env's pre_exec commands
        base = '%s/env/rp_named_env.%s' % (self._pwd, env_name)
        src  = '%s.env'                 %  base
        tgt  = '%s.%s.sh'               % (base, self.name.lower())

        # if the env does not yet exists - create
        # FIXME: this would need some file locking for concurrent executors. or
        #        add self._uid to path name
        if not os.path.isfile(tgt):

            src_env = ru.env_read(src)
            for ep in self.get_env_preserved():
                if ep in src_env:
                    src_env[ep] = f'${ep}:{src_env[ep]}'

            blacklist = self.get_env_blacklist()
            self._log.debug_5('blacklist: %s', blacklist)

            ru.env_prep(environment=src_env,
                        blacklist=blacklist,
                        unset=list(os.environ.keys()),
                        script_path=tgt)

        return tgt


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def get_exec(self, task):

        td           = task['description']
        task_exec    = td['executable']
        task_args    = td['arguments']
        task_argstr  = self._create_arg_string(task_args)
        command      = '%s %s' % (task_exec, task_argstr)

        return command.rstrip()


    # --------------------------------------------------------------------------
    #
    def get_partition_ids(self):

        # by default, launchers will only support a single partition
        return [0]


    # --------------------------------------------------------------------------
    #
    def _create_arg_string(self, args):

        if args:
            return ' '.join([ru.sh_quote(arg) for arg in args])
        else:
            return ''


    # --------------------------------------------------------------------------
    #
    def _get_mpi_info(self, exe):
        '''
        returns version and flavor of MPI version.
        '''

        version, flavor = None, None

        for info_opts in ['-V', '--version', '-info']:
            out, _, ret = ru.sh_callout('%s %s' % (exe, info_opts))
            if not ret:
                break

        if ret:
            raise RuntimeError('cannot identify MPI flavor [%s]' % exe)

        if 'pals' in exe.lower():
            flavor = self.MPI_FLAVOR_PALS

        for line in out.splitlines():

            if 'intel(r) mpi library for linux' in line.lower():
                # Intel MPI is hydra based
                version = line.split(',')[1].strip()
                flavor = self.MPI_FLAVOR_HYDRA

            elif 'hydra build details:' in line.lower():
                version = line.split(':', 1)[1].strip()
                flavor  = self.MPI_FLAVOR_HYDRA

            elif 'mvapich2' in line.lower():
                version = line.strip()
                flavor  = self.MPI_FLAVOR_HYDRA

            elif '(open mpi)' in line.lower():
                version = line.split(')', 1)[1].strip()
                flavor  = self.MPI_FLAVOR_OMPI

            elif 'ibm spectrum mpi' in line.lower():
                version = line.split(')', 1)[1].strip()
                flavor  = self.MPI_FLAVOR_SPECTRUM

            elif 'version' in line.lower():
                version = line.lower().split('version')[1].\
                          replace(':', '').strip()
                if not flavor:
                    flavor = self.MPI_FLAVOR_OMPI

            if version:
                break

        if not flavor:
            flavor = self.MPI_FLAVOR_UNKNOWN

        self._log.debug('mpi details [%s]: %s', exe, out)
        self._log.debug('mpi version: %s [%s]', version, flavor)

        return version, flavor


    # --------------------------------------------------------------------------
    #
    def get_env_blacklist(self):
        '''
        return a list of environment variable names which the launcher needs to
        ensure are *not* overwritten by the execution script or mechanism.
        Specifically, we need to make sure that any applied `named_env` will not
        set these variables.

        Individual launchers may extend this list with additional names.
        '''

        return [
                # RP internal variables
                'RP_*',

                # SANDBOX location
                'PWD',

                # resource details
                'ROCR_VISIBLE_DEVICES',
                'CUDA_VISIBLE_DEVICES',
        ]

    def get_env_preserved(self):
        '''List of environment variables that will be preserved. If a
        corresponding environment variable will be reassigned, then we extend
        its value with the original one (e.g., ENV_VAR=$ENV_VAR:new_value)'''
        return []


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2013-2020, http://radical.rutgers.edu'
__license__   = 'MIT'

import copy
import time

import threading as mt

from collections import defaultdict

import radical.utils as ru

from ...   import states as rps
from ..    import LaunchMethod
from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
#
class Flux(AgentExecutingComponent) :

    _shell = ru.which('bash') or '/bin/sh'

    # --------------------------------------------------------------------------
    #
    def initialize(self):
        '''
        This components has 3 strands of activity (threads):

          - the main thread listens for incoming tasks from the scheduler, and
            pushes them toward the watcher thread;
          - an event listener thread listens for flux events which signify task
            state updates, and pushes those events also to the watcher thread;
          - the watcher thread matches events and tasks, enacts state updates,
            and pushes completed tasks toward output staging.

        NOTE: we get tasks in *AGENT_SCHEDULING* state, and enact all
              further state changes in this component.
        '''

        super().initialize()

        # translate Flux states to RP states
        # NOTE: only map to states for which output channels are defined
        self._event_map = {
                         # 'submit'   : rps.AGENT_SCHEDULING,
                         # 'depend'   : rps.AGENT_SCHEDULING_PENDING,
                         # 'alloc'    : rps.AGENT_EXECUTING_PENDING,
                         # 'start'    : rps.AGENT_EXECUTING,
                           'cleanup'  : None,
                           'finish'   : rps.AGENT_STAGING_OUTPUT_PENDING,
                         # 'release'  : 'unschedule',
                           'free'     : None,
                           'clean'    : None,
                           'priority' : None,
                           'exception': rps.FAILED,
                          }

        lm_cfg = self.session.rcfg.launch_methods.get('FLUX')
        lm_cfg['pid']       = self.session.cfg.pid
        lm_cfg['reg_addr']  = self.session.cfg.reg_addr
        self._lm            = LaunchMethod.create('FLUX', lm_cfg, self._rm.info,
                                                  self._log, self._prof)

        # local state management
        self._tasks       = dict()             # task_id -> task
        self._events      = defaultdict(list)  # flux_id -> [events]
        self._events_lock = mt.Lock()

        self._task_count = 0

        # we only start the flux backend here
        self._lm.start_flux(event_cb=self._handle_event_cb)


    # --------------------------------------------------------------------------
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def _lm_fail_task(self, tasks, state):
        '''
        This will be called by the launch method in case a task error was
        detected during launch, before handing the task to flux
        '''

        for task in tasks:
            self._log.error('flux launch failed for %s: %s', task['uid'], state)

        self.advance_tasks(tasks, state=rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def get_task(self, tid):

        return self._tasks.get(tid)


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, task):

        # FIXME: clarify how to cancel tasks in Flux
        self._lm.cancel_task(task)


    # --------------------------------------------------------------------------
    #
    def _create_spec(self, task):

        tid = task['uid']

        self._log.debug('create spec: %s', tid)

        td     = task['description']
        uid    = task['uid']
        sbox   = task['task_sandbox_path']
        stdout = td.get('stdout') or '%s/%s.out' % (sbox, uid)
        stderr = td.get('stderr') or '%s/%s.err' % (sbox, uid)

        td['sandbox'] = sbox
        td['stdout']  = stdout
        td['stderr']  = stderr

        task['stdout'] = ''
        task['stderr'] = ''

        task['stdout_file'] = stdout
        task['stderr_file'] = stderr

        self._prof.prof('task_create_exec_start', uid=uid)
        _, exec_path = self._create_exec_script(self._lm, task)
        self._prof.prof('task_create_exec_ok', uid=uid)

        spec_dict = copy.deepcopy(td)
        spec_dict['uid']        = uid
        spec_dict['executable'] = '/bin/sh'
        spec_dict['arguments']  = ['-c', exec_path]

        self._prof.prof('task_to_spec_start', uid=uid)
        ret = ru.flux.spec_from_dict(spec_dict)
        self._prof.prof('task_to_spec_stop', uid=uid)

        return ret


    # --------------------------------------------------------------------------
    #
    def _handle_event_cb(self, task_id, event):

        ename = event.name
        state = None
        push  = True

        self._log.debug('flux event: %s: %s', task_id, ename)
        self._log.debug_3('          : %s', str(event.context))

        task = self._tasks.get(task_id)

        # handle some special events, fallback to _event_map otherwise
        if ename == 'start':
            # start task timeout handling, no further action
            self.handle_timeout(task)

        elif ename == 'unschedule':
            # free task resources, no further action
            self._prof.prof('unschedule_start', uid=task['uid'])
            self._prof.prof('unschedule_stop',  uid=task['uid'])  # ?

        elif ename == 'lm_failed':
            self._log.error('flux launch failed for %s', task_id)
            state = rps.FAILED
            push  = False

        elif ename == 'exception' and \
             event.context['type'] in ['cancel', 'timeout']:

            state = rps.AGENT_STAGING_OUTPUT_PENDING
            task['target_state'] = rps.CANCELED

        else:
            state = self._event_map.get(ename)


        if state is None:
            # no further state handling needed
            return

        self._log.debug('flux event mapped: %s -> %s', ename, state)

        # handle some states specifically
        if state == rps.AGENT_STAGING_OUTPUT_PENDING:

            if not task.get('target_state'):
                task['exit_code'] = event.context.get('status', 1)
                if task['exit_code']: task['target_state'] = rps.FAILED
                else                : task['target_state'] = rps.DONE

            # FIXME: run post-launch commands here.  Alas, this is
            #        synchronous, and thus potentially rather slow.
            tid  = task['uid']
            cmds = task['description'].get('post_launch')
            if cmds:
                for cmd in cmds:
                    self._log.debug('post-launch %s: %s', task['uid'], cmd)
                    out, err, ret = ru.sh_callout(cmd, shell=True,
                                                  cwd=task['task_sandbox_path'])
                    self._log.debug('post-launch %s: %s [%s][%s]',
                                                             tid, ret, out, err)
                    if ret:
                        self.advance_tasks(task, rps.FAILED,
                                           publish=True, push=False)
                        return

        # otherwise we just advance to the found state
        self.advance_tasks(task, state, ts=event.timestamp,
                           publish=True, push=push)


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance(tasks, rps.AGENT_EXECUTING, publish=True, push=False)

        n_parts = self._rm.info.n_partitions
        parts   = defaultdict(dict)  # part_id -> {task_id: task}

        for task in tasks:

            tid = task['uid']
            self._tasks[tid] = task

            try:
                # FIXME: pre_launch commands are synchronous and thus
                #        potentially slow.
                self._prof.prof('work_0', uid=tid)

                cmds = task['description'].get('pre_launch')
                if cmds:
                    sbox = task['task_sandbox_path']
                    ru.rec_makedir(sbox)

                    for cmd in cmds:
                        self._log.debug('pre-launch %s: %s', task['uid'], cmd)
                        out, err, ret = ru.sh_callout(cmd, shell=True,
                                              cwd=task['task_sandbox_path'])
                        self._log.debug('pre-launch %s: %s [%s][%s]',
                                        tid, ret, out, err)

                        if ret:
                            raise RuntimeError('cmd failed: %s' % cmd)

                self._prof.prof('work_1', uid=task['uid'])
                part_id = task['description']['partition']
                if part_id is None:
                    part_id = self._task_count % n_parts
                    self._task_count += 1

                self._prof.prof('work_2', uid=task['uid'])
                self._log.debug('task %s on partition %s', task['uid'], part_id)
                task['description']['environment']['RP_PARTITION_ID'] = part_id

                parts[part_id][tid] = self._create_spec(task)
                self._prof.prof('work_3', uid=task['uid'])

            except:
                self._log.exception('LM flux submit failed for %s', tid)
                self.advance_tasks(task, rps.FAILED, publish=True, push=False)

        self._lm.submit_tasks(parts)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import io
import os
import sys
import time
import queue


import radical.utils   as ru
import radical.pilot   as rp

import multiprocessing as mp
import threading       as mt

import dragon                      # pylint: disable=unused-import, import-error
mp.set_start_method("dragon")

from dragon.native.process       import Popen
from dragon.native.process       import Process
from dragon.native.process       import ProcessTemplate
from dragon.native.process_group import ProcessGroup


# ------------------------------------------------------------------------------
#
def get_stdio(conn) -> str:

    data = ''
    try:
        while True:
            data += conn.recv()

    except EOFError:
        return data

    finally:
        conn.close()


# ------------------------------------------------------------------------------
#
class Handler(object):
    '''
    Abstract handler to watch for dragon things completing.  Implementations
    will wrap around dragon's `ProcessGroup`s and
    `mp.Pool.apply_async.AsyncResult` objects.
    '''

    def is_alive(self):
        raise NotImplementedError('is_alive not implemented')

    def get_results(self):
        raise NotImplementedError('get_results not implemented')


# ------------------------------------------------------------------------------
#
class AsyncResultHandler(Handler):

    def __init__(self, res):
        self._res = res

    def is_alive(self):
        return not self._res.ready()


    # --------------------------------------------------------------------------
    #
    def get_results(self):

        assert not self.is_alive(), 'task still running'

        out = ''
        err = ''
        ret = 0
        val = None
        exc = None

        try:
            out, err, ret, val, exc = self._res.get()

        except Exception as e:
            err = repr(e)
            exc = repr(e)
            ret = -1

        return out, err, ret, val, exc


# ------------------------------------------------------------------------------
#
class GroupHandler(Handler):

    # --------------------------------------------------------------------------
    #
    def __init__(self, group):
        self._group = group

    def is_alive(self):
        return not self._group.inactive_puids


    # --------------------------------------------------------------------------
    #
    def get_results(self):

        assert not self.is_alive(), 'group still running'

        out = ''
        err = ''
        ret = 0
        val = None
        exc = None

        try:

            for item in self._group.inactive_puids:

                proc = Process(None, ident=item[0])
                out  += get_stdio(proc.stdout_conn)
                err  += get_stdio(proc.stderr_conn)

                if item[1] != 0:
                    ret = item[1]

            self._group.close()
            self._group = None

        except Exception as e:
            err = repr(e)
            exc = e
            ret = -1

        return out, err, ret, val, exc


# ------------------------------------------------------------------------------
#
class Server(object):

    # FIXME: profile events

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        self._uid = ru.generate_id('radical.pilot.dragon')

        self._log  = ru.Logger(self._uid)
        self._prof = ru.Profiler(self._uid, path='.')
        self._term = mt.Event()

        self._log.info('start dragon executor server %s', self._uid)

        self._pwatcher = ru.PWatcher(uid='%s.pw' % self._uid, log=self._log)
        self._pwatcher.watch(os.getpid())
        self._pwatcher.watch(os.getppid())

        self._pipe_in  = None
        self._pipe_out = None

        self._pool  = None
      # self._slots = mp.cpu_count()
        self._slots = int(sys.argv[1])
        self._free  = self._slots

        self._watcher_queue = queue.Queue()
        self._logger_queue  = mp.Queue()

        self._logger = mt.Thread(target=self._logger_thread)
        self._logger.daemon = True
        self._logger.start()

        self._watcher = mt.Thread(target=self._watcher_thread)
        self._watcher_event  = mt.Event()
        self._watcher.daemon = True
        self._watcher.start()

        self._worker = mt.Thread(target=self._worker_thread)
        self._worker_event  = mt.Event()
        self._worker.daemon = True
        self._worker.start()

        self._worker_event.wait(timeout=5)
        self._watcher_event.wait(timeout=5)

        if not self._worker_event.is_set():
            raise RuntimeError('launcher thread did not start')

        if not self._watcher_event.is_set():
            raise RuntimeError('watcher thread did not start')

        assert self._pipe_in,  'pipe_in not set'
        assert self._pipe_out, 'pipe_out not set'

        url_in  = ru.Url(ru.as_string(self._pipe_in.url))
        url_out = ru.Url(ru.as_string(self._pipe_out.url))

        url_in.host  = ru.get_hostip()
        url_out.host = ru.get_hostip()

        msg = 'ZMQ_ENDPOINTS %s %s\n' % (str(url_in), str(url_out))
        self._log.debug('URLs: %s', msg)
        sys.stdout.write(msg)
        sys.stdout.flush()

        # run forever
        while not self._term.is_set():
            time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def _worker_thread(self):
        '''
        handle incoming requests from the parent process

        The requests are expected to be dictionaries with the following keys:
            - cmd:  the command to execute (run, cancel, stop)
            - task: the task to execute the command on
        '''

        self._log.info('start launcher')

        self._pipe_in  = ru.zmq.Pipe(ru.zmq.MODE_PULL)
        self._worker_event.set()

        self._pool = mp.Pool(self._slots)

        try:
            while True:

                msg = self._pipe_in.get_nowait(1)

                if not msg:
                    continue

                cmd = msg.get('cmd')

                if cmd == 'stop':
                    self._log.info('stop')
                    self._term.set()
                    sys.exit(0)

                elif cmd == 'cancel':
                    self._log.error('task cancellation not yet implemented')

                elif cmd == 'run':
                    self._launch(msg['task'])

                else:
                    raise ValueError('unknown command %s' % cmd)


        except Exception as e:
            self._log.exception('error in main loop: %s', e)
            raise

        finally:
            self._term.set()


    # --------------------------------------------------------------------------
    #
    def _launch(self, task):

        tid   = task['uid']
        ranks = task['description']['ranks']
        mode  = task['description']['mode']

        # we may need to wait for free slots.
        # FIXME: check GPUs, multithreadind, mem etc.
        while self._free < ranks:
            self._log.debug_9('wait for slots for %s [%d] - %d free', tid,
                              ranks, self._free)
            time.sleep(1)

        self._free -= ranks
        self._log.debug('found slots for %s [%d] - %d free', tid, ranks,
                        self._free)


        if mode == rp.TASK_EXECUTABLE:
            if ranks > 1:
                return self._launch_mpi(task)
            else:
                return self._launch_nonmpi(task)
        else:
            return self._launch_function(task)


    # --------------------------------------------------------------------------
    #
    def _launch_mpi(self, task):

        self._log.debug('task %s: launch', task['uid'])

        tid   = task['uid']
        sbox  = task['task_sandbox_path']
        ranks = task['description']['ranks']
        exe   = task['exec_path']
        out   = task['stdout_file']
        err   = task['stderr_file']

        cmd   = '/bin/sh'
        args  = ['-c', '%s >> %s 2>> %s' % (exe, out, err)]
        env   = os.environ.copy()

        self._log.debug('task %s: [%d ranks]: %s %s', tid, ranks, cmd, args)

        # TODO: use placement policies for GPU tasks
        group = ProcessGroup(restart=False, pmi_enabled=True, policy=None)

        for rank in range(ranks):

            env['DRAGON_RANK'] = str(rank)
            tmp = ProcessTemplate(target=cmd, args=args, env=env, cwd=sbox,
                                  stdout=Popen.PIPE, stderr=Popen.PIPE,
                                  stdin=Popen.DEVNULL)
            group.add_process(nproc=1, template=tmp)

        group.init()
        group.start()

        handler = GroupHandler(group)

        task['dragon_handler'] = (handler, ranks)

        self._log.debug('task %s: dragon handler %s', tid, handler)
        self._watcher_queue.put(task)


    # --------------------------------------------------------------------------
    #
    def _launch_nonmpi(self, task):

        self._log.debug('task %s: nonmpi launch', task['uid'])

        ranks = task['description']['ranks']
        assert ranks == 1, 'non-MPI tasks must have ranks==1'

        res = self._pool.apply_async(self._run_nonmpi,
                                     [task, self._logger_queue])

        handler = AsyncResultHandler(res)

        task['dragon_handler'] = (handler, ranks)
        self._watcher_queue.put(task)


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _run_nonmpi(task, lq):

        val = None
        exc = None
        out = ''
        err = ''
        ret = 0

        tid = task['uid']
        exe = task['exec_path']

        lq.put(('debug', 'task %s: launch', tid))

        try:
            # TODO: use placement policies for GPU tasks
            out, err, ret = ru.sh_callout('/bin/sh %s' % exe, shell=True)
            lq.put(('debug', 'task %s: done: %s : %s : %s', tid, out, err, ret))

        except Exception as e:
            lq.put(('exception', 'task %s: call failed: %s', tid, e))
            exc = repr(e)
            ret = 1
            err += '\ncall failed: %s' % e

        return out, err, ret, val, exc


    # --------------------------------------------------------------------------
    #
    def _launch_function(self, task):

        self._log.debug('task %s: launch func', task['uid'])

        ranks = task['description']['ranks']

        assert ranks == 1, 'function tasks must have ranks==1'

        res = self._pool.apply_async(self._run_function, [task,
                                                          self._logger_queue])
        handler = AsyncResultHandler(res)

        task['dragon_handler'] = (handler, ranks)
        self._watcher_queue.put(task)


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _run_function(task, lq):

        tid    = task['uid']
        func   = task['description']['function']
        args   = task['description']['args']
        kwargs = task['description']['kwargs']

        lq.put(('debug', 'task %s: run function', tid))

        # FIXME: populate sbox?
      # sbox   = task['task_sandbox_path']
      # out    = task['stdout_file']
      # err    = task['stderr_file']

        # check if `func_name` is a global name
        names   = dict(list(globals().items()) + list(locals().items()))
        to_call = names.get(func)

        # check if we have a serialized object
        if not to_call:
            lq.put(('debug', 'function serialized: %d: %s', len(func), func))

            try:
                to_call, _args, _kwargs = rp.PythonTask.get_func_attr(func)

            except Exception:
                lq.put('exception', 'function is not a PythonTask [%s] ', tid)

            else:
                if args or kwargs:
                    raise ValueError('`args` and `kwargs` must be empty for '
                                     'PythonTask function [%s]' % tid)
                else:
                    args   = _args
                    kwargs = _kwargs

        if not to_call:
            lq.put(('debug', 'no %s in \n%s\n', func, names))
            raise ValueError('%s callable %s not found: %s' % (tid, func, task))

        # make sure we capture stdout / stderr
        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        # set the task environment
        # FIXME

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            lq.put(('prof', 'rank_start', tid, time.time()))
            lq.put(('debug', 'to call %s: %s : %s', to_call, args, kwargs))

            val = to_call(*args, **kwargs)

            lq.put(('prof', 'rank_stop', tid, time.time()))
            out = strout.getvalue()
            err = strerr.getvalue()
            exc = (None, None)
            ret = 0

        except Exception as e:
            lq.put(('exception', '_call failed: %s', tid, e))
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\ncall failed: %s' % e)
            exc = repr(e)
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr

            # FIXME
            # os.environ = old_env

        lq.put(('debug', '%s: got "%s"', tid, out))

        return out, err, ret, val, exc


    # --------------------------------------------------------------------------
    #
    def _watcher_thread(self):

        try:

            to_collect = dict()

            self._log.info('start watcher')

            self._pipe_out = ru.zmq.Pipe(ru.zmq.MODE_PUSH)
            self._log.debug('pipe_out PUSH created: %s - %s',
                            ru.as_string(self._pipe_out.url), ru.get_hostip())
            self._watcher_event.set()

            time.sleep(0.5)

            self._pipe_out.put({'cmd': 'hello'})
            self._log.debug('pipe_out sent hello')

            # get completed tasks
            while True:

              # self._log.debug('watcher loop: %s', list(to_collect.keys()))

                # FIXME: get_nowait
                while not self._watcher_queue.empty():

                    task  = self._watcher_queue.get()
                    tid   = task['uid']
                    to_collect[tid] = task

                    self._log.debug('task %s: watch', tid)

                if not to_collect:
                    time.sleep(0.1)
                 #  self._log.debug('no tasks to collect')
                    continue

                collected = list()
                for tid,task in to_collect.items():

                    handler, _ = task['dragon_handler']
                  # self._log.debug('task %s: check handler %s', tid, handler)

                    if handler.is_alive():
                        pass

                    else:
                        del task['dragon_handler']

                        ranks = task['description']['ranks']
                        self._free += ranks

                        self._log.debug('task %s: collect [%d] - %d free',
                                        tid, ranks, self._free)

                        out, err, ret, val, exc = handler.get_results()

                        self._log.debug('task %s: completed %s : %s : %s : %s : %s',
                                        tid, out, err, ret, val, exc)

                        task['stdout']       = out
                        task['stderr']       = err
                        task['exit_code']    = ret
                        task['return_value'] = val
                        task['exception']    = str(exc)

                        if ret == 0: task['target_state'] = rp.DONE
                        else       : task['target_state'] = rp.FAILED

                        collected.append(tid)

                     #  self._log.debug('push %s: completed %s', tid, task)
                        self._pipe_out.put({'cmd': 'done',
                                            'task': task})
                     #  self._log.debug('pushed %s', tid)

                for tid in collected:
                    del to_collect[tid]
                  # self._log.debug('deleted %s', tid)

                # avoid busy loop
                if not collected:
                    time.sleep(0.5)

        except:
            self._log.exception('error in watcher thread')
            raise


    # --------------------------------------------------------------------------
    #
    def _logger_thread(self):

        while True:

            vals = self._logger_queue.get()

            mode = vals[0]
            if mode == 'prof':
                self._prof.prof(vals[1], uid=vals[2], ts=vals[3])

            elif mode == 'debug':
                self._log.debug(vals[1], *vals[2:])

            elif mode == 'exception':
                self._log.exception(vals[1], *vals[2:])

            else:
                self._log.error('unknown logger mode %s: %s', mode, vals)


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    s = Server()


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2024, The RADICAL-Cybertools Team'
__license__   = 'MIT'


import os
import time

from typing import List

from rc.process import Process

import threading     as mt

import radical.utils as ru

from ...  import states    as rps
from ...  import constants as rpc

from .popen import Popen


# ------------------------------------------------------------------------------
#
class Dragon(Popen):

    # --------------------------------------------------------------------------
    #
    def initialize(self):
        '''
        This method is called by the base class during component
        initialization.

        It will start the dragon executor. The executor will communicate the endpoint
        URLs back to us. We will then use those endpoints to communicate with the dragon
        process.
        '''

        super().initialize()

        self._url_in    = None
        self._url_out   = None
        self._start_evt = mt.Event()

        # ----------------------------------------------------------------------
        #
        def line_cb(proc: Process, lines : List[str]) -> None:
            for line in lines:
                self._log.info('line: %s', line)
                if line.startswith('ZMQ_ENDPOINTS '):
                    _, self._url_out, self._url_in = line.split()
                    self._start_evt.set()
                    break

        def state_cb(proc: Process, state: str):
            self._log.debug('process state: %s' % state)
            if state == 'failed':
                self._log.error('dragon executor failed')
                self._log.error('stdout: %s', proc.stdout)
                self._log.error('stderr: %s', proc.stderr)
                self._start_evt.set()
        # ----------------------------------------------------------------------

        n_nodes = len(self._rm.info.node_list)
        n_slots = self._rm.info.cores_per_node * n_nodes

        cmd  = ''
        if os.path.exists('/opt/cray/libfabric/1.22.0/lib64'):
            cmd += 'dragon-config '
            cmd += '-a "ofi-runtime-lib=/opt/cray/libfabric/1.22.0/lib64"; '

        cmd += 'dragon -N %d ' % n_nodes
        cmd += '-l ERROR '
        cmd += 'radical-pilot-dragon-executor.py %d' % n_slots

        self._log.debug('=== cmd: %s', cmd)

        p = Process(cmd, shell=True)
        p.register_cb(p.CB_OUT_LINE, line_cb)
        p.register_cb(p.CB_STATE, state_cb)
        p.polldelay = 0.1
        p.start()

        self._pwatcher = ru.PWatcher(uid='%s.pw' % self._uid, log=self._log)
        self._pwatcher.watch(os.getpid())
        self._pwatcher.watch(p.pid)

        self._start_evt.wait()

        self._log.debug('dragon eps: %s - %s', self._url_in, self._url_out)

        assert self._url_in
        assert self._url_out

        self._pipe_out = ru.zmq.Pipe(ru.zmq.MODE_PUSH, self._url_out)

        # run watcher thread
        self._watcher = mt.Thread(target=self._dragon_watch, args=[self._url_in])
        self._watcher.daemon = True
        self._watcher.start()


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task):

        # before we start handling the task, check if it should run in a named
        # env.  If so, inject the activation of that env in the task's pre_exec
        # directives.
        tid  = task['uid']
        td   = task['description']
        sbox = task['task_sandbox_path']

        # prepare stdout/stderr
        task['stdout'] = ''
        task['stderr'] = ''

        stdout_file    = td.get('stdout') or '%s.out' % tid
        stderr_file    = td.get('stderr') or '%s.err' % tid

        if stdout_file[0] != '/':
            task['stdout_file']       = '%s/%s' % (sbox, stdout_file)
            task['stdout_file_short'] = '$RP_TASK_SANDBOX/%s' % stdout_file
        else:
            task['stdout_file']       = stdout_file
            task['stdout_file_short'] = stdout_file

        if stderr_file[0] != '/':
            task['stderr_file']       = '%s/%s' % (sbox, stderr_file)
            task['stderr_file_short'] = '$RP_TASK_SANDBOX/%s' % stderr_file
        else:
            task['stderr_file']       = stderr_file
            task['stderr_file_short'] = stderr_file

        launcher, lname = self._rm.find_launcher(task)

        if not launcher:
            raise RuntimeError('no launcher found for %s' % task)

        task['launcher_name'] = lname

        _, exec_fullpath = self._create_exec_script(launcher, task)
        task['exec_path']   = exec_fullpath

        # send task to dragon
        self._log.debug('launch task %s', task['uid'])

        if False:
            task['target_state'] = rps.DONE
            task['exit_code']    = 0

            self._prof.prof('unschedule_start', uid=tid)

            self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, [task])

            self.advance([task], rps.AGENT_STAGING_OUTPUT_PENDING,
                                 publish=True, push=True)

        else:
            self._pipe_out.put({'cmd': 'run', 'task': task})
            self._log.debug('sent task %s', task['uid'])

        # handle task timeout if needed
        self.handle_timeout(task)


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, task):
        '''
        This method is called by the base class to actually
        cancel the task.
        '''

        # send cancel request to dragon
        self._pipe_out.put({'cmd': 'cancel', 'uid': task['uid']})


    # --------------------------------------------------------------------------
    #
    def _dragon_watch(self, url_in):
        '''
        This method watches the dragon endpoint for task completion messages.

        :param url_in:  the dragon endpoint to watch
        '''

        self._log.debug('dragon watch: %s', url_in)

        pipe_in  = ru.zmq.Pipe(ru.zmq.MODE_PULL, url_in)
        self._log.debug('pipe_in  PUll created: %s - %s',
                            ru.as_string(pipe_in.url), ru.get_hostname())
        time.sleep(0.5)

        try:
            while not self._term.is_set():

              # self._log.debug('pipe_in  receive')
                msg = pipe_in.get_nowait(5.1)

                if not msg:
                  # self._log.info('pipe_in  no msg')
                    continue

                cmd = msg.get('cmd')

                self._log.debug('pipe_in  receive [%s]', cmd)

                if cmd == 'hello':
                    self._log.info('pipe_in  hello pipe')
                    continue

                if cmd != 'done':
                    raise ValueError('unsupported command %s' % cmd)

                task = msg['task']
                tid  = task['uid']

                self._prof.prof('unschedule_start', uid=tid)
                self._log.debug('dragon watch: %s : %s', cmd, tid)

                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, [task])

                self.advance([task], rps.AGENT_STAGING_OUTPUT_PENDING,
                                     publish=True, push=True)

        except Exception as e:
            self._log.exception('Error in ExecWorker watch loop (%s)' % e)
            self.stop()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"

import time
import queue

from collections import defaultdict
from functools   import partial

import threading       as mt
import multiprocessing as mp

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class Flux(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    class Partition(object):

        def __init__(self):

            self.uid      = None
            self.uri      = None
            self.service  = None
            self.helper   = None

    class Event(ru.TypedDict):
        _schema = {'name'     : str,
                   'timestamp': float}


    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, session, prof):

        self._partitions  = list()
        self._idmap       = dict()             # flux_id -> task_id
        self._part_map    = dict()             # task_id -> part_id
        self._events      = defaultdict(list)  # flux_id -> [events]
        self._events_lock = mt.Lock()
        self._in_queues   = list()
        self._out_queues  = list()

        super().__init__(name, lm_cfg, rm_info, session, prof)


    # --------------------------------------------------------------------------
    #
    @property
    def partitions(self):
        return self._partitions


    def can_launch(self, task):
        raise RuntimeError('method cannot be used on Flux LM')


    def get_launch_cmds(self, task, exec_path):
        raise RuntimeError('method cannot be used on Flux LM')


    def get_launcher_env(self):
        raise RuntimeError('method cannot be used on Flux LM')


    def get_rank_cmd(self):
        return 'export RP_RANK=$FLUX_TASK_RANK\n'


    def get_env_blacklist(self):
        '''
        extend the base blacklist by FLUX related environment variables
        '''

        ret = super().get_env_blacklist()

        ret.extend([
                    'FLUX_*',
                    'MPICH_*',
                    'PMI_*',
                    'PMIX_*',
                    # Slurm included since we launch Flux using Srun LM
                    'SLURM_*'
                   ])

        return ret

    def get_env_preserved(self):
        ret = super().get_env_preserved()
        ret.extend([
            'LD_LIBRARY_PATH'
        ])
        return ret


    # --------------------------------------------------------------------------
    #
    def _terminate(self):

        for part in self._partitions:

            self._log.debug('stop partition %s [%s]', part.uid, part.uri)

            try   : part.service.stop()
            except: pass

            for fh in part.helper:
                try   : fh.stop()
                except: pass

        self._partitions = list()


    # --------------------------------------------------------------------------
    #
    def init_from_scratch(self, env, env_sh):

        # just make sure we can load flux

        fm = ru.flux.FluxModule()
        if fm.exc:
            raise RuntimeError('flux import failed: %s' % fm.exc)

        self._log.debug('flux core   : %s', fm.core)
        self._log.debug('flux job    : %s', fm.job)
        self._log.debug('flux exe    : %s', fm.exe)
        self._log.debug('flux version: %s', fm.version)

        lm_info = {'env'       : env,
                   'env_sh'    : env_sh}

        return lm_info


    # --------------------------------------------------------------------------
    #
    def init_from_info(self, lm_info):

        pass


    # --------------------------------------------------------------------------
    #
    def start_flux(self, event_cb):

        self._event_cb = event_cb

        # start one flux instance per partition.  Flux instances are hosted in
        # partition threads which are fed tasks through queues.
        self._prof.prof('flux_start')

        # first partition nodelist into, well, partitions
        partitions = ru.partition(self._rm_info.node_list,
                                  self._rm_info.n_partitions)

        # run one partition thread per partition
        for part_id, nodes in enumerate(partitions):

            self._log.debug('starting flux partition %d on %d nodes',
                            part_id, len(nodes))

            q_in  = mp.Queue()
            q_out = mp.Queue()
            part_proc = mp.Process(target=self._part_thread,
                                     args=(part_id, nodes, q_in, q_out),
                                     name='rp.flux.part.%d' % part_id)
            part_proc.start()
            # FIXME: add process watcher

            self._in_queues.append(q_in)
            self._out_queues.append(q_out)

        for q_out in self._out_queues:

            # wait for all partition threads to finish starting
            start = time.time()
            while True:
                now = time.time()
                if now - start > 600:
                    raise RuntimeError('flux partition did not start')

                try:
                    val = q_out.get(timeout=1)
                except queue.Empty:
                    continue

                if val is not True:
                    raise RuntimeError('flux partition startup failed')

                break

            self._log.debug('flux partition started')

        # start watcher thread
        watcher = mt.Thread(target=self._queue_watcher)
        watcher.daemon = True
        watcher.start()

        self._prof.prof('flux_start_ok')


    # --------------------------------------------------------------------------
    #
    def _part_thread(self, part_id, nodes, q_in, q_out):

        threads_per_node = self._rm_info.cores_per_node  # == hw threads
        gpus_per_node    = self._rm_info.gpus_per_node

        self._log.info('starting partition thread %d on %d nodes',
                       part_id, len(nodes))

        part = self.Partition()

        self._log.debug('flux partition %d starting', part_id)

        # FIXME: this is a hack for frontier and will only work for slurm
        #        resources.  If Flux is to be used more widely, we need to
        #        pull the launch command from the agent's resource manager.
        launcher = None
        srun     = ru.which('srun')
        mpiexec  = ru.which('mpiexec')
        nodelist = ','.join([node['name'] for node in nodes])
        if srun:
            launcher = 'srun --nodes %d --nodelist %s --ntasks-per-node 1 ' \
                       '--cpus-per-task=%d --gpus-per-task=%d ' \
                       '--export=ALL' \
                       % (len(nodes), nodelist, threads_per_node, gpus_per_node)

        elif mpiexec:
            launcher = 'mpiexec -n %d --host %s' % (len(nodes), nodelist)

        self._log.debug('flux launcher: %s', launcher)

        part.service = ru.FluxService(launcher=launcher)
        part.service.start()

        if not part.service.ready(timeout=600):
            raise RuntimeError('flux service did not start')

        part.uri = part.service.r_uri

        self._log.debug('flux service %s: %s', part.service.uid, part.uri)

        part.helper = ru.FluxHelper(uri=part.uri)
        part.helper.start()

        partial_cb = partial(self._job_event_cb, q_out)
        part.helper.register_cb(partial_cb)

        self._log.debug('flux helper: %s', part.helper.uid)

        # signal startup
        q_out.put(True)


        # we now wait for tasks to submit to our flux service
        while True:

            try:
                msg = q_in.get(timeout=1)

            except queue.Empty:
                continue

            cmd = msg[0]

            if cmd == 'cancel':
                tid = msg[1]
                self._log.debug('cancel task %s on partition %d', tid, part_id)
                part.helper.cancel(tid)
                continue

            assert cmd == 'submit', cmd

            try:
                tasks = msg[1]
                specs = list()

                for tid in tasks:

                    self._prof.prof('part_0', uid=tid)
                    specs.append(tasks[tid])

                for tid in tasks:
                    self._prof.prof('submit_0', uid=tid)

                fids = part.helper.submit(specs)
                for fid, tid in zip(fids, tasks.keys()):
                    self._prof.prof('submit_1', uid=tid)
                  # self._log.debug('push flux job id: %s -> %s', tid, fid)
                    q_out.put(['job_id', (tid, fid)])

                self._log.debug('%s: submitted %d tasks', part.uid, len(tasks))

            except Exception:
                self._log.exception('LM flux submit failed')
                for tid in tasks:
                    self._event_cb(tid, self.Event(name='lm_failed',
                                                   timestamp=time.time()))



    # --------------------------------------------------------------------------
    #
    def submit_tasks(self, parts):

        for part_id, specs in parts.items():

            q_in = self._in_queues[part_id]
            q_in.put(['submit', specs])

            for tid in specs:
                self._log.debug('register %s on part %d', tid, part_id)
                self._part_map[tid] = part_id


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, task):

        tid     = task['uid']
        part_id = self._part_map.get(tid)

        if part_id is None:
            self._log.error('cancel: no part for task %s', tid)
            return

        self._log.debug('cancel task %s on partition %d', tid, part_id)

        q_in = self._in_queues[part_id]
        q_in.put(['cancel', tid])


    # --------------------------------------------------------------------------
    #
    def _queue_watcher(self):

        while True:

            busy = False
            for q_out in self._out_queues:

                if q_out.empty():
                    continue

                busy = True
                cmd, event = q_out.get()

              # self._log.debug('=== pull event: %s: %s', cmd, event)

                if cmd == 'job_id':
                    task_id, flux_id = event
                    self._job_id_handler(task_id, flux_id)

                elif cmd == 'event':
                    flux_id, event = event
                    self._job_event_handler(flux_id, event)

                else:
                    self._log.error('unknown flux event: %s', cmd)

            if not busy:
                time.sleep(0.1)


    # --------------------------------------------------------------------------
    #
    def _job_id_handler(self, task_id, flux_id):

        with self._events_lock:

            self._log.debug('flux job id: %s -> %s', task_id, flux_id)

            self._idmap[flux_id] = task_id

            events = self._events.get(flux_id, [])

            self._log.debug('flux job id: %s -> %d events', task_id, len(events))
            if events:
                del self._events[flux_id]


        for event in events:
            self._log.debug('job id event: %s, event: %s', task_id, event.name)
            self._event_cb(task_id, event)


    # --------------------------------------------------------------------------
    #
    def _job_event_cb(self, q_out, flux_id, event):

      # self._log.debug('=== push flux job event: %s: %s', flux_id, event.name)
        q_out.put(['event', (flux_id, event)])


    # --------------------------------------------------------------------------
    #
    def _job_event_handler(self, flux_id, event):

        self._log.debug('flux event: %s: %s', flux_id, event.name)

        with self._events_lock:

            task_id = self._idmap.get(flux_id)

            if not task_id:
                self._log.info('no task id for flux job %s: %s', flux_id,
                                event.name)
                self._events[flux_id].append(event)
                return

        self._event_cb(task_id, event)


# ------------------------------------------------------------------------------




__copyright__ = 'Copyright 2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import dill
import pickle
import codecs
import tempfile


_obj_dir       = tempfile.gettempdir()
_obj_file_name = 'rp_obj.pkl'
_obj_file_path = os.path.join(_obj_dir, _obj_file_name)


# ------------------------------------------------------------------------------
# Custom Exceptions
# ------------------------------------------------------------------------------

class SerializationError(Exception):
    def __init__(self, message: str, original_exception: Exception):
        super().__init__(f"{message}: {str(original_exception)}")
        self.original_exception = original_exception


class DeserializationError(Exception):
    def __init__(self, message: str, original_exception: Exception):
        super().__init__(f"{message}: {str(original_exception)}")
        self.original_exception = original_exception


# ------------------------------------------------------------------------------
# Serialization
# ------------------------------------------------------------------------------

def serialize_obj(obj):
    '''
    Serialize object using dill.
    '''
    try:
        if callable(obj):
            return dill.dumps(obj)
        else:
            return dill.dumps(obj, recurse=True)
    except Exception as e:
        try:
            return dill.dumps(obj, byref=True)
        except Exception as e2:
            raise SerializationError("Failed to serialize object", e2) from e2


def serialize_file(obj, fname=None):
    '''
    Serialize object to file.
    '''
    if not fname:
        fname = _obj_file_path

    try:
        with open(fname, 'wb') as f:
            f.write(serialize_obj(obj))
        return fname
    except Exception as e:
        raise SerializationError("Failed to serialize object to file", e) from e


def serialize_bson(obj):
    '''
    Serialize object to base64-encoded BSON.
    '''
    try:
        return codecs.encode(pickle.dumps(obj), "base64").decode()
    except Exception as e:
        raise SerializationError("Failed to serialize object to BSON", e) from e


# ------------------------------------------------------------------------------
# Deserialization
# ------------------------------------------------------------------------------

def deserialize_file(fname):
    '''
    Deserialize object from file.
    '''
    try:
        with open(fname, 'rb') as f:
            return dill.load(f)
    except Exception as e:
        raise DeserializationError("Failed to deserialize object from file", e) from e


def deserialize_obj(data):
    '''
    Deserialize object from bytes.
    '''
    try:
        return dill.loads(data)
    except Exception as e:
        raise DeserializationError("Failed to deserialize object from data", e) from e


def deserialize_bson(obj):
    '''
    Deserialize object from base64-encoded BSON.
    '''
    try:
        return pickle.loads(codecs.decode(obj.encode(), "base64"))
    except Exception as e:
        raise DeserializationError("Failed to deserialize object from BSON", e) from e
